package cli

import (
	"fmt"
	"os"
	"runtime"
	"strings"

	"git.urbach.dev/cli/q/src/build"
	"git.urbach.dev/cli/q/src/config"
)

// Build parses the arguments and creates a build.
func Build(args []string) int {
	_, err := buildExecutable(args)

	if err != nil {
		fmt.Fprintln(os.Stderr, err)
		return exit(err)
	}

	return 0
}

// buildExecutable creates a new build with the given arguments.
func buildExecutable(args []string) (*build.Build, error) {
	b := build.New()

	for i := 0; i < len(args); i++ {
		switch args[i] {
		case "--assembly", "-a":
			config.ShowAssembly = true

		case "--dry":
			config.Dry = true

		case "--sections":
			config.Sections = true

		case "--statistics":
			config.ShowStatistics = true

		case "--verbose", "-v":
			config.ShowAssembly = true
			config.ShowStatistics = true

		case "--arch":
			i++

			if i >= len(args) {
				return b, &ExpectedParameterError{Parameter: "arch"}
			}

			switch args[i] {
			case "arm":
				config.SetTargetArch(config.ARM)
			case "x86":
				config.SetTargetArch(config.X86)
			default:
				return b, &InvalidValueError{Value: args[i], Parameter: "arch"}
			}

		case "--os":
			i++

			if i >= len(args) {
				return b, &ExpectedParameterError{Parameter: "os"}
			}

			switch args[i] {
			case "linux":
				config.TargetOS = config.Linux
			case "mac":
				config.TargetOS = config.Mac
			case "windows":
				config.TargetOS = config.Windows
			default:
				return b, &InvalidValueError{Value: args[i], Parameter: "os"}
			}

		default:
			if strings.HasPrefix(args[i], "-") {
				return b, &UnknownParameterError{Parameter: args[i]}
			}

			b.Files = append(b.Files, args[i])
		}
	}

	if config.TargetOS == config.UnknownOS {
		return b, &InvalidValueError{Value: runtime.GOOS, Parameter: "os"}
	}

	if config.TargetArch == config.UnknownArch {
		return b, &InvalidValueError{Value: runtime.GOARCH, Parameter: "arch"}
	}

	if len(b.Files) == 0 {
		b.Files = append(b.Files, ".")
	}

	err := start(b)
	return b, err
}

// start starts the build by running the compiler and then writing the result to disk.
func start(b *build.Build) error {
	result, err := b.Run()

	if err != nil {
		return err
	}

	if config.ShowAssembly {
		result.PrintInstructions()
	}

	if config.ShowStatistics {
		result.PrintStatistics()
	}

	if config.Dry {
		return nil
	}

	return result.WriteFile(b.Executable())
}
